package guitplugin;

import com.guit.client.apt.GwtEvent;
import com.guit.client.apt.GwtEvent.EventKind;
import com.sun.mirror.apt.AnnotationProcessor;
import com.sun.mirror.apt.AnnotationProcessorEnvironment;
import com.sun.mirror.apt.Filer;
import com.sun.mirror.declaration.AnnotationTypeDeclaration;
import com.sun.mirror.declaration.ClassDeclaration;
import com.sun.mirror.declaration.Declaration;
import com.sun.mirror.declaration.FieldDeclaration;

import java.io.IOException;
import java.io.PrintWriter;
import java.util.Collection;

public class EventProcessor implements AnnotationProcessor {

    AnnotationProcessorEnvironment env;

    public EventProcessor(AnnotationProcessorEnvironment env) {
        this.env = env;
    }

    public AnnotationProcessorEnvironment getEnvironment() {
        return env;
    }

    public void process() {
        Filer filer = env.getFiler();

        Collection<Declaration> annotatedTypes = env.getDeclarationsAnnotatedWith((AnnotationTypeDeclaration) env.getTypeDeclaration(EventProcessorFactory.annotations.get(0)));
        for (Declaration decl : annotatedTypes) {
            if (decl instanceof ClassDeclaration) {
                ClassDeclaration classDeclaration = (ClassDeclaration) decl;
                try {
                    Collection<FieldDeclaration> fields = classDeclaration.getFields();

                    EventKind kind = decl.getAnnotation(GwtEvent.class).value();

                    String packageName;
                    switch (kind) {
                        case DOM:
                            packageName = "com.google.gwt.event.dom.client";
                            break;
                        case SHARED:
                            packageName = "com.google.gwt.event.logical.shared";
                            break;
                        default:
                            packageName = classDeclaration.getPackage().getQualifiedName();
                            break;
                    }

                    String simpleName = classDeclaration.getSimpleName();
                    String eventName = simpleName + "Event";
                    String handlerName = simpleName + "Handler";

                    PrintWriter writer = filer.createSourceFile(packageName + "." + eventName);
                    writer.println("package " + packageName + ";");
                    writer.println("import com.google.gwt.event.shared.GwtEvent;");
                    writer.println("import com.google.gwt.event.shared.EventHandler;");
                    writer.println("import " + packageName + "." + eventName + "." + handlerName + ";");

                    writer.println("public class " + eventName + " extends GwtEvent<" + handlerName + "> {");

                    StringBuilder constructorParameters = new StringBuilder();
                    StringBuilder asignaments = new StringBuilder();
                    StringBuilder toString = new StringBuilder();
                    for (FieldDeclaration f : fields) {
                        String type = f.getType().toString();
                        String name = f.getSimpleName();
                        
                        if (toString.length() > 0) {
                        	toString.append(", ");
        				}
                        toString.append(name + "=" + "\" + " + name + " + \"");

                        if (constructorParameters.length() > 0) {
                            constructorParameters.append(", ");
                        }

                        asignaments.append("this." + name + " = " + name + ";");

                        constructorParameters.append(type);
                        constructorParameters.append(" ");
                        constructorParameters.append(name);

                        writer.println("  private final " + type + " " + name + ";");

                        writer.println("  public " + type + " get" + name.substring(0, 1).toUpperCase() + name.substring(1) + "() {");
                        writer.println("    return " + name + ";");
                        writer.println("  }");
                    }
                    
                    writer.println("  public String toString() {");
        			writer.println("      return \"" + eventName + "[" + toString.toString() + "]\";");
        			writer.println("  }");
        			
        			writer.println("  public String toDebugString() {");
        			writer.println("      return toString();");
        			writer.println("  }");

                    writer.println("  public static interface " + handlerName + " extends EventHandler {");
                    writer.println("    void on" + simpleName + "(" + eventName + " event);");
                    writer.println("  }");

                    writer.println("  public final static Type<" + handlerName + "> TYPE = new Type<" + handlerName + ">();");

                    writer.println("  public " + eventName + "(" + constructorParameters.toString() + ") {");
                    writer.println("  " + asignaments.toString());
                    writer.println("  }");

                    writer.println("  @Override");
                    writer.println("  protected void dispatch(" + handlerName + " handler) {");
                    writer.println("    handler.on" + simpleName + "(this);");
                    writer.println("  }");

                    writer.println("  @Override");
                    writer.println("  public Type<" + handlerName + "> getAssociatedType() {");
                    writer.println("    return TYPE;");
                    writer.println("  }");
                    writer.println("}");
                    writer.close();
                } catch (IOException e) {
                    throw new IllegalStateException(e);
                }
            }
        }
    }
}
