package guitplugin;

import java.io.IOException;
import java.io.PrintWriter;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map.Entry;

import com.guit.client.apt.Cache;
import com.guit.client.apt.GuitService;
import com.guit.client.apt.Ignore;
import com.sun.mirror.apt.AnnotationProcessor;
import com.sun.mirror.apt.AnnotationProcessorEnvironment;
import com.sun.mirror.apt.Filer;
import com.sun.mirror.declaration.AnnotationTypeDeclaration;
import com.sun.mirror.declaration.InterfaceDeclaration;
import com.sun.mirror.declaration.MethodDeclaration;
import com.sun.mirror.declaration.ParameterDeclaration;
import com.sun.mirror.type.TypeMirror;
import com.sun.mirror.type.VoidType;
import com.sun.mirror.util.DeclarationFilter;

public class GuitServiceProcessor implements AnnotationProcessor {

	AnnotationProcessorEnvironment env;

	public GuitServiceProcessor(AnnotationProcessorEnvironment env) {
		this.env = env;
	}

	public AnnotationProcessorEnvironment getEnvironment() {
		return env;
	}

	@Override
	public void process() {
		Filer filer = env.getFiler();

		for (InterfaceDeclaration d : DeclarationFilter
				.getFilter(InterfaceDeclaration.class)
				.filter(env.getDeclarationsAnnotatedWith((AnnotationTypeDeclaration) env
						.getTypeDeclaration(GuitService.class
								.getCanonicalName())),
						InterfaceDeclaration.class)) {

			String dpackage = d.getPackage().getQualifiedName();
			if (!d.getPackage().getSimpleName().equals("server")) {
				env.getMessager().printError(d.getPosition(),
						"All services must be on a 'server' package");
				return;
			}

			String actionPkg = dpackage.substring(0, dpackage.lastIndexOf("."))
					+ ".client.action";
			String hanlderPkg = dpackage + ".handler";

			Collection<? extends MethodDeclaration> methods = d.getMethods();
			for (MethodDeclaration m : methods) {
				if (m.getAnnotation(Ignore.class) == null) {
					String methodName = m.getSimpleName();
					methodName = methodName.substring(0, 1).toUpperCase()
							+ methodName.substring(1);
					String simpleName = methodName;
					String handlerClass = dpackage + ".handler." + methodName
							+ "Handler";
					HashMap<String, String> responseFields = new HashMap<String, String>();
					HashMap<String, String> actionFields = new HashMap<String, String>();

					StringBuilder gettersList = new StringBuilder();

					Collection<ParameterDeclaration> parameters = m
							.getParameters();
					for (ParameterDeclaration p : parameters) {
						String pName = p.getSimpleName();
						actionFields.put(pName, p.getType().toString());

						if (gettersList.length() > 0) {
							gettersList.append(", ");
						}
						pName = pName.substring(0, 1).toUpperCase()
								+ pName.substring(1);
						gettersList.append("action.get" + pName + "()");
					}

					TypeMirror returnType = m.getReturnType();
					boolean doesntReturnsVoid = returnType != null
							&& !(returnType instanceof VoidType);
					if (doesntReturnsVoid) {
						responseFields.put("result", returnType.toString());
					}
					generateAction(actionFields, responseFields, filer,
							handlerClass, simpleName, actionPkg,
							m.getAnnotation(Cache.class) != null);

					PrintWriter writer;
					try {
						writer = filer.createSourceFile(hanlderPkg + "."
								+ simpleName + "Handler");
					} catch (IOException e) {
						throw new RuntimeException(e);
					}

					String service = d.getQualifiedName();

					writer.println("package " + hanlderPkg + ";");
					writer.println();

					writer.println("import com.guit.client.command.action.CommandException;                            ");
					writer.println("import com.guit.client.command.action.Handler;                                     ");
					writer.println("import " + actionPkg + "." + simpleName
							+ "Action;                         ");
					writer.println("import " + actionPkg + "." + simpleName
							+ "Action." + simpleName + "Response;     ");
					writer.println("import " + service
							+ ";                                  ");
					writer.println("                                                                                   ");
					writer.println("import com.google.inject.Inject;                                                        ");
					writer.println("                                                                                   ");
					writer.println("public class "
							+ simpleName
							+ "Handler implements                                         ");
					writer.println("        Handler<" + simpleName + "Action, "
							+ simpleName
							+ "Response> {                          ");
					writer.println("                                                                                   ");
					writer.println("    @Inject                                                                        ");
					writer.println("    "
							+ d.getSimpleName()
							+ " service;                                               ");
					writer.println("                                                                                   ");
					writer.println("    @Override                                                                      ");
					writer.println("    public " + simpleName
							+ "Response handle(" + simpleName
							+ "Action action)                    ");
					writer.println("            throws CommandException {                                              ");
					if (doesntReturnsVoid) {
						writer.println("        return new " + simpleName
								+ "Response(service." + m.getSimpleName() + "("
								+ gettersList + "));          ");
					} else {
						writer.println("        service." + m.getSimpleName()
								+ "(" + gettersList + ");          ");
						writer.println("        return new " + simpleName
								+ "Response();          ");
					}
					writer.println("    }                                                                              ");
					writer.println("}                                                                                  ");
					writer.close();
				}
			}
		}
	}

	public static void generateAction(HashMap<String, String> actionFields,
			HashMap<String, String> responseFields, Filer filer,
			String handlerClass, String simpleName, String packageName,
			boolean cacheable) {
		try {
			String actionName = simpleName + "Action";
			String responseName = simpleName + "Response";

			PrintWriter writer = filer.createSourceFile(packageName + "."
					+ actionName);
			writer.println("package " + packageName + ";");

			if (cacheable) {
				writer.println("import com.guit.client.command.action.Cacheable;");
			}

			writer.println("import com.guit.client.command.HashCodeBuilder;");
			writer.println("import com.guit.client.command.action.Action;");
			writer.println("import com.guit.client.command.action.ActionHandler;");
			writer.println("import com.guit.client.command.action.Response;");
			writer.println("import com.guit.client.GuitEntryPoint;");
			writer.println("import com.guit.client.async.AbstractAsyncCallback;");

			writer.println("import " + packageName + "." + actionName + "."
					+ responseName + ";");

			writer.println("@ActionHandler(\"" + handlerClass + "\")");
			writer.print("public class " + actionName + " implements Action<"
					+ responseName + ">");

			// Cache
			if (cacheable) {
				writer.print(", Cacheable ");
			}

			writer.println(" {");

			StringBuilder constructor = new StringBuilder();
			StringBuilder assignments = new StringBuilder();

			StringBuilder actionString = new StringBuilder();

			for (Entry<String, String> f : actionFields.entrySet()) {
				String type = f.getValue();
				String name = f.getKey();

				if (constructor.length() > 0) {
					constructor.append(", ");
				}
				constructor.append(type + " " + name);
				assignments.append("this." + name + " = " + name + ";");

				printGetterAndSetter(writer, type, name, actionName);

				if (actionString.length() > 0) {
					actionString.append(", ");
				}
				actionString.append(name + ": " + "\" + " + name + " + \"");
			}

			writer.println("  public " + actionName + "("
					+ constructor.toString() + ") {");
			writer.println("    " + assignments.toString());
			writer.println("  }");

			writer.println("  public boolean equals(Object o) {");
			writer.println("    if (this == null) return false;");
			writer.println("    if (this == o) return true;");
			writer.println("    if (o instanceof "
					+ actionName
					+ ") { return o.hashCode() == hashCode(); } else { return super.equals(o);}");
			writer.println("  }");

			writer.println("  public int hashCode() {");
			writer.print("    return new HashCodeBuilder()");
			for (Entry<String, String> f : actionFields.entrySet()) {
				String name = f.getKey();
				writer.print(".append(" + name + ")");
			}
			writer.println(".toHashCode();");
			writer.println("  }");

			if (constructor.length() > 0) {
				writer.println("  public " + actionName + "() {");
				writer.println("  }");
			}

			writer.println("  public String toString() {");
			writer.println("      return \"" + actionString.toString() + "\";");
			writer.println("  }");

			writer.println("  public static " + actionName
					+ " create() { return new " + actionName + "(); }");

			writer.println("  public void execute(AbstractAsyncCallback<"
					+ responseName + "> callback) {");
			writer.println("      GuitEntryPoint.getCommandService().execute(this, callback);");
			writer.println("  }");

			writer.println("  public void execute() {");
			writer.println("      GuitEntryPoint.getCommandService().execute(this, new AbstractAsyncCallback<"
					+ responseName + ">() {");
			writer.println("          public void success(" + responseName
					+ " response) {");
			writer.println("          }");
			writer.println("      });");
			writer.println("  }");

			writer.println("  public void executeLater(AbstractAsyncCallback<"
					+ responseName + "> callback) {");
			writer.println("      GuitEntryPoint.getCommandService().executeLater(this, callback);");
			writer.println("  }");

			writer.println("  public void executeLater() {");
			writer.println("      GuitEntryPoint.getCommandService().executeLater(this, new AbstractAsyncCallback<"
					+ responseName + ">() {");
			writer.println("          public void success(" + responseName
					+ " response) {");
			writer.println("          }");
			writer.println("      });");
			writer.println("  }");

			writer.println("  public static class " + responseName
					+ " implements Response {");

			constructor = new StringBuilder();
			assignments = new StringBuilder();

			StringBuilder responseString = new StringBuilder();
			for (Entry<String, String> f : responseFields.entrySet()) {
				String type = f.getValue();
				String name = f.getKey();

				if (constructor.length() > 0) {
					constructor.append(", ");
				}
				constructor.append(type + " " + name);
				assignments.append("this." + name + " = " + name + ";");

				printGetterAndSetter(writer, type, name, responseName);

				if (responseString.length() > 0) {
					responseString.append(", ");
				}
				responseString.append(name + ": " + "\" + " + name + " + \"");
			}

			writer.println("  public String toString() {");
			writer.println("      return \"" + responseString.toString()
					+ "\";");
			writer.println("  }");

			writer.println("  public " + responseName + "("
					+ constructor.toString() + ") {");
			writer.println("    " + assignments.toString());
			writer.println("  }");

			if (constructor.length() > 0) {
				writer.println("  public " + responseName + "() {");
				writer.println("  }");
			}

			writer.println("  }");

			writer.println("}");
			writer.close();
		} catch (IOException e) {
			throw new RuntimeException(e);
		}
	}

	private static void printGetterAndSetter(PrintWriter writer, String type,
			String name, String className) {
		writer.println("  private " + type + " " + name + ";");

		writer.println("  public " + type + " get"
				+ name.substring(0, 1).toUpperCase() + name.substring(1)
				+ "() {");
		writer.println("    return " + name + ";");
		writer.println("  }");

		writer.println("  public void set" + name.substring(0, 1).toUpperCase()
				+ name.substring(1) + "(" + type + " " + name + ") {");
		writer.println("    this." + name + " = " + name + ";");
		writer.println("  }");

		writer.println("  public " + className + " " + name + "(" + type + " "
				+ name + ") {");
		writer.println("    this." + name + " = " + name + ";");
		writer.println("    return this;");
		writer.println("  }");
	}
}
